import platform

import websockets

from hvm.chains.base import (
    BaseChain,
)
from hp2p.events import (
    PeerCountRequest,
    PeerCountResponse,
)
from hp2p.service import (
    BaseService,
)
from helios import (
    __version__,
)
from helios.extensibility import (
    PluginContext,
)
from helios.constants import (
    SYNC_LIGHT,
)
from helios.plugins.builtin.light_peer_chain_bridge.light_peer_chain_bridge import (
    EventBusLightPeerChain,
)
from helios.utils.db_proxy import (
    create_db_manager,
)
from helios.utils.version import (
    construct_helios_client_identifier,
)

from helios.plugins.builtin.ethstats.ethstats_client import (
    EthstatsClient,
    EthstatsMessage,
    EthstatsData,
    timestamp_ms,
)


class EthstatsService(BaseService):
    def __init__(
        self,

        context: PluginContext,

        server_url: str,
        server_secret: str,
        node_id: str,
        node_contact: str,
    ) -> None:
        super().__init__()

        self.context = context

        self.server_url = server_url
        self.server_secret = server_secret
        self.node_id = node_id
        self.node_contact = node_contact

        self.chain = self.get_chain()

    async def _run(self) -> None:
        while self.is_operational:
            try:
                self.logger.info('Connecting to %s...' % self.server_url)
                async with websockets.connect(self.server_url) as websocket:
                    client: EthstatsClient = EthstatsClient(
                        websocket,
                        self.node_id,
                        token=self.cancel_token,
                    )

                    self.run_child_service(client)

                    await self.wait_first(
                        self.server_handler(client),
                        self.statistics_handler(client),
                    )
            except websockets.ConnectionClosed as e:
                self.logger.info('Connection to %s is closed: %s' % (self.server_url, e))

            self.logger.info('Reconnecting in 5s...')
            await self.sleep(5)

    # Wait for messages from server, respond when they arrive
    async def server_handler(self, client: EthstatsClient) -> None:
        while self.is_operational:
            message: EthstatsMessage = await client.recv()

            if message.command == 'node-pong':
                await client.send_latency((timestamp_ms() - message.data['clientTime']) // 2)
            elif message.command == 'history':
                # TODO: send actual history
                pass
            else:
                self.logger.info('Server message received')

    # Periodically send statistics and ping server to calculate latency
    async def statistics_handler(self, client: EthstatsClient) -> None:
        await client.send_hello(self.server_secret, self.get_node_info())

        while self.is_operational:
            await client.send_node_ping()
            await client.send_stats(await self.get_node_stats())
            await client.send_block(self.get_node_block())

            await self.sleep(5)

    def get_node_info(self) -> EthstatsData:
        return {
            'name': self.node_id,
            'contact': self.node_contact,
            'node': construct_helios_client_identifier(),
            'net': self.context.chain_config.network_id,
            'port': self.context.chain_config.port,
            'os': platform.system(),
            'os_v': platform.release(),
            'client': __version__,
            'canUpdateHistory': False,
        }

    def get_node_block(self) -> EthstatsData:
        head = self.chain.get_canonical_head()

        return {
            'number': head.block_number,
            'hash': head.hex_hash,
            'difficulty': head.difficulty,
            'totalDifficulty': self.chain.get_score(head.hash),
            'transactions': [],
            'uncles': [],
        }

    async def get_node_stats(self) -> EthstatsData:
        response: PeerCountResponse = await self.context.event_bus.request(
            PeerCountRequest()
        )

        return {
            'active': True,
            'peers': response.peer_count,
        }

    def get_chain(self) -> BaseChain:
        db_manager = create_db_manager(self.context.chain_config.database_ipc_path)
        db_manager.connect()

        chain_class = self.context.chain_config.node_class.chain_class

        if self.context.chain_config.sync_mode == SYNC_LIGHT:
            header_db = db_manager.get_headerdb()  # type: ignore
            chain = chain_class(
                header_db,
                peer_chain=EventBusLightPeerChain(self.context.event_bus)
            )
        else:
            db = db_manager.get_db()  # type: ignore
            chain = chain_class(db)

        return chain
